// ========================================================================
// vector3.h
//
// R^(3x1)
// ========================================================================

#ifndef	__VECTOR3_H
#define	__VECTOR3_H

#include	<math.h>

// | e1 |
// | e2 |
// | e3 |




// _̒`
#if FLOATISDOUBLE
typedef double VAR  ;
#else
typedef float VAR  ;
#endif

#define	INLINE inline

#define	PI (3.1415926535f)


//
//
struct vector3
{
friend struct vector4;
friend class matrix3x3;
friend class matrix4x4;
private:
	union{
		VAR   e[3];
		struct{
			VAR   x,y,z;
		};
	};


public:
	vector3() {}
	vector3(VAR   f){
		e[0]=f;e[1]=f;e[2]=f;
	}
	vector3(VAR   e0,VAR   e1,VAR   e2){
		e[0]=e0;e[1]=e1;e[2]=e2;
	}
	~vector3() {}


	// one term operator
	VAR  & operator[](int n){ return e[n]; }
	vector3 operator- (){ return vector3(-e[0],-e[1],-e[2]); }

	// two term operator
	vector3& operator= (VAR   f){ return *this=vector3(f); }
	vector3& operator+=(vector3& a);
	vector3& operator-=(vector3& a);
	vector3& operator*=(VAR   a);
	vector3& operator/=(VAR   a);

	friend vector3 operator+(vector3& a,vector3& b);
	friend vector3 operator-(vector3& a,vector3& b);
	friend VAR     operator*(vector3& a,vector3& b);	//scalar product
	friend vector3 operator*(vector3& a,VAR      b);
	friend vector3 operator*(VAR      a,vector3& b);
	friend vector3 operator/(vector3& a,VAR      b);

	friend bool operator==(vector3& a,vector3& b);
	friend bool operator==(vector3& a,VAR      b);
	friend bool operator==(VAR      a,vector3& b);
	friend bool operator!=(vector3& a,vector3& b);
	friend bool operator!=(vector3& a,VAR      b);
	friend bool operator!=(VAR      a,vector3& b);


	// 
	VAR      dot  (vector3& a);
	vector3  cross(vector3& a);
	vector3  indmul( vector3& a );	// independent multiply ( u.x*v.x , u.y*v.y , u.z*v.z )

	//
	VAR      norm();
	VAR      distance();
	vector3  unit();

	//
	void normalize();
	
	// ( [v1,0] [v2,0] [v3,0] ) cross product
	static VAR   cross(vector3& v1,vector3& v2,vector3& v3);

	// 

};





// ========================================================================


inline vector3& vector3::operator+=(vector3& a)
{
	e[0]+=a.e[0];
	e[1]+=a.e[1];
	e[2]+=a.e[2];
	return *this;
}

inline vector3& vector3::operator-=(vector3& a)
{
	e[0]-=a.e[0];
	e[1]-=a.e[1];
	e[2]-=a.e[2];
	return *this;
}

inline vector3& vector3::operator*=(VAR   a)
{
	return *this=*this * a;
}

inline vector3& vector3::operator/=(VAR   a)
{
	return *this=*this / a;
}


// ========================================================================


inline vector3 operator+(vector3& a,vector3& b)
{
	return vector3( a.e[0]+b.e[0], a.e[1]+b.e[1], a.e[2]+b.e[2] );
}


inline vector3 operator-(vector3& a,vector3& b)
{
	return vector3( a.e[0]-b.e[0], a.e[1]-b.e[1], a.e[2]-b.e[2] );
}

inline VAR   operator*(vector3& a,vector3& b)
{
	return a.e[0]*b.e[0] + a.e[1]*b.e[1] + a.e[2]*b.e[2];
}

inline vector3 operator*(vector3& a,VAR      b)
{
	return vector3( a.e[0]*b, a.e[1]*b, a.e[2]*b );
}

inline vector3 operator*(VAR      a,vector3& b)
{
	return b*a;
}

inline vector3 operator/(vector3& a,VAR      b)
{
	return vector3( a.e[0]/b, a.e[1]/b, a.e[2]/b );
}

// ========================================================================

inline bool operator==(vector3& a,vector3& b)
{
	return a.e[0]==b.e[0] && a.e[1]==b.e[1] && a.e[2]==b.e[2] ;
}

inline bool operator==(vector3& a,VAR      b)
{
	return a.e[0]==b && a.e[1]==b && a.e[2]==b ;
}

inline bool operator==(VAR      a,vector3& b)
{
	return a==b.e[0] && a==b.e[1] && a==b.e[2] ;
}

inline bool operator!=(vector3& a,vector3& b)
{
	return !(a==b);
}

inline bool operator!=(vector3& a,VAR      b)
{
	return !(a==b);
}

inline bool operator!=(VAR      a,vector3& b)
{
	return !(a==b);
}

// ========================================================================

inline VAR      vector3::dot  (vector3& a)
{
	return a*a;
}

inline vector3  vector3::cross(vector3& a)
{
	return  vector3(
			e[1]*a.e[2] -e[2]*a.e[1],
	        e[2]*a.e[0] -e[0]*a.e[2],
			e[0]*a.e[1] -e[1]*a.e[0]
			);
}			

inline VAR      vector3::norm()
{
	return *this * *this;
}

inline VAR      vector3::distance()
{
	return sqrt(norm());
}

inline vector3  vector3::unit()
{
	VAR   d=distance();
	if( d==0 ) return 0;
	return *this/d;
}

// ========================================================================


inline void vector3::normalize()
{
	*this=unit();
}

// ========================================================================

inline VAR   vector3::cross(vector3& v1,vector3& v2,vector3& v3)
{
	return
		+v1.e[2]*v2.e[1]*v3.e[0] -v1.e[1]*v2.e[2]*v3.e[0]
		-v1.e[2]*v2.e[0]*v3.e[1] +v1.e[0]*v2.e[2]*v3.e[1]
		+v1.e[1]*v2.e[0]*v3.e[2] -v1.e[0]*v2.e[1]*v3.e[2];
/*	return
		+v1.z*v2.y*v3.x -v1.y*v2.z*v3.x
		-v1.z*v2.x*v3.y +v1.x*v2.z*v3.y
		+v1.y*v2.x*v3.z -v1.x*v2.y*v3.z;
*/
}


inline vector3  vector3::indmul( vector3& a )
{
	return vector3( e[0]*a[0],e[1]*a[1],e[2]*a[2] );
}

#endif //__VECTOR3_H
