

#include <stdio.h>
#include <ctype.h>
#include "raytrace.h"

#ifdef WIN32
#define	WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif

using namespace raytrace;




class SceneLoader
{
public:
	SceneLoader( Render* r,Scene* s,View* v )
    {
		render=r;
		scene=s;
		view=v;
    }
	bool LoadText( const char * fname );
	
private:
	int getword( char* out );
	bool chkword( const char* str );
	bool tstword( const char* str );
	bool getvector( vector3* v );
	bool getvectorn( std::vector<int> * v );
	bool getquaternion( quaternion * q );
	bool getcolor( color* c );
	bool getfloat( double* d );
	bool getint( int* x );
	bool getstring( char* str );
	
	void LoadView();
	void LoadLight();

	Object* LoadObject();

	Object* LoadSphere();
	Object* LoadFloor();
	Object* LoadBox();
	Object* LoadTriangularPyramid();
	Object* LoadSimpleMesh();
	Object* LoadQuadricSurface();
	Object* LoadCSG();

	bool LoadExtendCommand( Object* p );
	bool LoadMaterial( Material* m );
	
	enum WORDTYPE {
		ENDOFFILE,
		SYMBOL,
		NAME,
		NUMBER
	};
	
private:
	Render *render;
	Scene  *scene;
	View   *view;
	
	FILE *fp;
	char line[256];
	char lastchar;
	int linenum;
	WORDTYPE wordtype;
	char word[256];
};

bool SceneLoader::getint( int * x ) 
{
	if( wordtype!=NUMBER ) return false;
	*x=atoi(word);
	getword(word);
	return true;
}

bool SceneLoader::getfloat( double * d ) 
{
	if( wordtype!=NUMBER ) return false;
	*d=atof(word);
	getword(word);
	return true;
}

bool SceneLoader::getvector( vector3 * v )
{
	if( !tstword("(") ) return false;
	double x,y,z;
	if( !getfloat(&x) ) throw "";
	tstword(",");
	if( !getfloat(&y) ) throw "";
	tstword(",");
	if( !getfloat(&z) ) throw "";
	*v=vector3(x,y,z);
	chkword(")");

	return true;
}

bool SceneLoader::getquaternion( quaternion * q )
{
	if( !tstword("(") ) return false;
	double w,x,y,z;
	if( !getfloat(&w) ) throw "";
	tstword(",");
	if( !getfloat(&x) ) throw "";
	tstword(",");
	if( !getfloat(&y) ) throw "";
	tstword(",");
	if( !getfloat(&z) ) throw "";
	*q=quaternion( w,x,y,z );
	chkword(")");

	return true;
}

bool SceneLoader::getcolor( color * c )
{
	if( !tstword("(") ) return false;
	double x,y,z;
	if( !getfloat(&x) ) throw "";
	tstword(",");
	if( !getfloat(&y) ) throw "";
	tstword(",");
	if( !getfloat(&z) ) throw "";
	c->Set(x,y,z);
	chkword(")");
	return true;
}

bool SceneLoader::getvectorn( std::vector<int> * v )
{
	if( !tstword("(") ) return false;
	while(true)
	{
		int x;
		if( !getint(&x) ) break;
		v->push_back(x);
	}
	chkword(")");
	return true;
}

bool SceneLoader::getstring( char* str )
{
	if( tstword("\"") ) {
		*str=0;
		while( !tstword("\"") ) {
			strcat( str,word );
			getword(word);
		}
		return true;
	}
	if( wordtype==ENDOFFILE || wordtype==SYMBOL ) return false;
	strcpy( str,word );
	getword(word);
	return true;
}


bool SceneLoader::tstword( const char * str )
{
	
	if( strcmp( word,str ) != 0 ) return false;
	getword(word);
	return true;
}

bool SceneLoader::chkword( const char * str ) 
{
	if( strcmp( word,str ) != 0 ) throw "";
	getword(word);
	return true;
}

int SceneLoader::getword( char *out )
{
	char c=lastchar;
	do {

		while(true) {
			if( c==EOF ){
				wordtype=ENDOFFILE;
				lastchar=c;
				return ENDOFFILE;
				//throw "end of file";
			}
			if( isspace(c) ) {
				if( c=='\n' ) linenum++;
				c=fgetc( fp );
			} else break;
		}
		
		
		// comment
		if( c=='/' ) {
			int cc=fgetc(fp);
			if( cc=='/' ){
				while( (c=fgetc(fp))!='\n' );
				continue;
			}
			lastchar=cc;
			out[0]=c;
			out[1]=0;
			return wordtype=SYMBOL;
		}
		
		// number
		if( ('0'<=c && c<='9' ) || c=='.' || c=='-' ) {
			do {
				*out++=c;
				c=fgetc(fp);
			} while( ('0'<=c && c<='9') || c=='.' || c=='-' || c=='e' );
			lastchar=c;
			*out++=0;
			return wordtype=NUMBER;
		}

		// symbol(kigou)
		if( ! (__iscsym(c) || isspace(c)) ) {
			out[0]=c;
			out[1]=0;
			lastchar=fgetc(fp);
			out[2]=0;
			return wordtype=SYMBOL;
		}
		
		// name
		if( __iscsymf(c) ) {
			do {
				*out++=c;
				c=fgetc(fp);
			} while( ('a'<=c && c<='z') || ('0'<=c && c<='9') );
			lastchar=c;
			*out++=0;
			return wordtype=NAME;
		}
		
		c=fgetc(fp);
	} while(1);
	
}

bool SceneLoader::LoadText( const char * fname )
{
	fp=fopen( fname,"rt" );
	
	if( ! fp ) return false;
	
	try
    {
		linenum=1;
		lastchar=fgetc(fp);
		getword(word);
		while( wordtype!=ENDOFFILE )
		{
			if( tstword("view") ) LoadView();
			else if( tstword("light") ) LoadLight();
			else if( tstword("object") ) {
				Object *p=LoadObject();
				scene->AddObject(p);
			} else if( tstword("#") ) {
				if( tstword("endoffile") ) break;
				else throw "not found this extra command";
			}
			else getword(word);

		}
    }
	catch( char* e )
    {
		if( wordtype!=ENDOFFILE ) {
			printf( "%3d:%s",linenum,e );
			__asm int 3;
		}
    }
	
	
	fclose(fp);
	return true;
}

void SceneLoader::LoadView()
{
	vector3 eye,up,dir;
	double x,y;
	
	chkword("{");
	
	getvector( &eye );	tstword(",");
	getvector( &dir );	tstword(",");
	getvector( &up );	tstword(",");
	getfloat( &x );		tstword(",");
	getfloat( &y );
	
	chkword("}");

	view->SetView( eye,(dir-eye).unit(),-up,x,y );
}

void SceneLoader::LoadLight()
{
	vector3 pos;
	color c;
	VAR damp=0;
	int rep1=1,rep2=1;
	vector3 dir1=0,dir2=0,dir=vector3(1,1,1);
	VAR ndir=0,dircos=0.8;

	if( tstword("line") ) {
		getint(&rep1);
		getvector(&dir1);
	}

	if( tstword("area") ) {
		getint(&rep1);
		getint(&rep2);
		getvector(&dir1);
		getvector(&dir2);
	}
	
	chkword("{");

	getvector( &pos ); tstword(",");
	getcolor( &c );    tstword(",");
	getfloat( &damp ); tstword(",");
	getvector( &dir );	tstword(",");	// w
	getfloat( &ndir );tstword(",");
	getfloat( &dircos );

	chkword("}");
	
	for( int x2=0;x2<rep2;x2++ )
	{
		for( int x1=0;x1<rep1;x1++ )
		{
			Light *p=new Light( c/(double)(rep1*rep2),pos + (0.5+x1-0.5*rep1)*dir1 + (0.5+x2-0.5*rep2 )*dir2, damp,dir,ndir,dircos );
			scene->AddLight( p );
		}
	}
	
}

Object* SceneLoader::LoadObject()
{
	Object *p;
	if( tstword("sphere") ) p=LoadSphere();
	else if( tstword("floor") ) p=LoadFloor();
	else if( tstword("box") ) p=LoadBox();
	else if( tstword("tripyramid") ) p=LoadTriangularPyramid();
	else if( tstword("simplemesh") ) p=LoadSimpleMesh();
	else if( tstword("quadricsurface") ) p=LoadQuadricSurface();
	else if( tstword("csg") ) p=LoadCSG();
	else throw "must set object type";
	return p;
}

bool SceneLoader::LoadMaterial( Material * m )
{
	if( ! tstword( "material" ) ) return false;

	color diffuse,specolor,ambient=0.1,opaque=0;
	double specular,reflect,trans,refract;

	chkword("{");

	getcolor( &diffuse );	tstword(",");
	getcolor( &specolor );	tstword(",");
	getfloat( &specular );	tstword(",");
	getfloat( &reflect );	tstword(",");
	getfloat( &trans );		tstword(",");
	getfloat( &refract );	tstword(",");
	getcolor( &ambient );	tstword(",");
	getcolor( &opaque );

	chkword("}");

	*m = Material( diffuse,specolor,specular,reflect,trans,refract,ambient,opaque );
	return true;
}

bool SceneLoader::LoadExtendCommand( Object* p )
{
	bool b=false;
	Material m;
	while(true)
	{
		if( LoadMaterial(&m) ) {
			p->ExtendCommand( "material",&m );
		} else if( tstword("texture") ) {
			chkword("{");
			char fname[256];
			getstring( fname );
			p->ExtendCommand( "texture",fname );
			chkword("}");
		} else break;
		b=true;
	}
	return b;
}

Object* SceneLoader::LoadSphere()
{
	chkword("{");
	vector3 pos;
	double radius;
	
	getvector( &pos ); 	tstword(",");
	getfloat( &radius ); tstword(",");
	
	Sphere *p=new Sphere(pos,radius);

	LoadExtendCommand(p);

	chkword("}");
	return p;
}

Object* SceneLoader::LoadFloor()
{
	chkword("{");
	vector3 pos,normal;
	color c1,c2;
	double size;
	
	getvector( &pos ); 	tstword(",");
	getvector( &normal );	tstword(",");
	getcolor( &c1 ); 	tstword(",");
	getcolor( &c2 ); 	tstword(",");
	getfloat( &size );	tstword(",");

	Floor *p=new Floor( pos,normal,c1,c2,size );
	LoadExtendCommand(p);

	chkword("}");
	return p;
}

Object* SceneLoader::LoadBox()
{
	chkword("{");
	vector3 pos;
	quaternion q;
	double width,height,depth;
	
	getvector( &pos ); 	tstword(",");
	getquaternion( &q );tstword(",");
	getfloat( &width );	tstword(",");
	getfloat( &height );tstword(",");
	getfloat( &depth );	tstword(",");
	
	SimpleMesh *p=new SimpleMesh();
	p->SetBox( pos,q,width,height,depth );
	LoadExtendCommand(p);
	
	chkword("}");
	return p;
}

Object* SceneLoader::LoadTriangularPyramid()
{
	chkword("{");
	vector3 pos;
	quaternion q;
	double width,height,depth;
	
	getvector( &pos ); 	tstword(",");
	getquaternion( &q );tstword(",");
	getfloat( &width );	tstword(",");
	getfloat( &height );tstword(",");
	getfloat( &depth );	tstword(",");
	
	SimpleMesh *p=new SimpleMesh();
	p->SetTriangularPyramid( pos,q,width,height );
	LoadExtendCommand(p);

	chkword("}");
	return p;
}

Object* SceneLoader::LoadSimpleMesh()
{
	chkword("{");
	vector3 pos;
	quaternion q;
	double sizex,sizey,sizez;

	SimpleMesh *p=new SimpleMesh;

	getvector( &pos ); 	tstword(",");
	getquaternion( &q );tstword(",");
	getfloat( &sizex );	tstword(",");
	getfloat( &sizey );	tstword(",");
	getfloat( &sizez );

	while(true)
	{
		tstword(",");
		if( tstword("index") )
		{
			chkword("{");
			while(true) {
				std::vector<int> index;
				if( !getvectorn( &index ) ) break;
				p->AddSurface( index.size(),&index[0] );
			}
			chkword("}");
		} else
		if( tstword("vertex") )
		{
			chkword("{");
			std::vector<vector3> vertex;
			while(true) {
				vector3 v;
				if( !getvector( &v ) ) break;
				vertex.push_back(v);
			}
			p->SetVertices( vertex.size(),&vertex[0] );
			chkword("}");
		}
		else if( LoadExtendCommand(p) ) {}
		else break;

	}

	chkword("}");

	p->SetPos( pos );
	p->SetQ( q );
	p->RecalcPos();
	return p;
}

Object* SceneLoader::LoadQuadricSurface()
{
	chkword("{");
	double a[10];
	for( int i=0;i<10;i++ )
		getfloat( &a[i] );	tstword(",");
	QuadricSurface *p=new QuadricSurface( a );
	LoadExtendCommand(p);
	chkword("}");
	return p;
}

Object* SceneLoader::LoadCSG()
{
	CSGTYPE t=CSG_OR;
	if( tstword("and") ) t=CSG_AND;
	else if( tstword("or") ) t=CSG_OR;
	else if( tstword("sub") ) t=CSG_SUB;
	chkword("{");
	ObjectCSG *p = new ObjectCSG;

	chkword("src1");
	Object *src1=LoadObject();

	chkword("src2");
	Object *src2=LoadObject();

	p->SetType( t );
	p->SetObject( src1,src2 );

	chkword("}");
	return p;
}

void SaveBmp( void* _data, int width,int height,int bpp )
{
	BYTE *data=(BYTE*)_data;
	int filesize=sizeof BITMAPFILEHEADER + sizeof BITMAPINFOHEADER + width*height*3;
	BYTE *pBuf = new BYTE [ filesize ];
	ZeroMemory( pBuf,filesize );

	LPBITMAPFILEHEADER pHead=(LPBITMAPFILEHEADER)pBuf;
	LPBITMAPINFOHEADER pInfo=(LPBITMAPINFOHEADER)(pBuf+sizeof BITMAPFILEHEADER);

	pHead->bfType='M'*256+'B';
	pHead->bfSize=filesize;
	pHead->bfOffBits=sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFOHEADER);

	pInfo->biSize=sizeof BITMAPINFOHEADER;
	pInfo->biWidth=width;
	pInfo->biHeight=height;
	pInfo->biPlanes=1;
	pInfo->biBitCount=24;

	BYTE *pDat=((BYTE*)pInfo)+sizeof BITMAPINFOHEADER;

	if( bpp==32 )
	{
		for( int y=height-1;y>=0;y-- ) {
			for( int x=0;x<width;x++ ) {
				*pDat++ = data[ (x+y*width)*4 + 0 ];
				*pDat++ = data[ (x+y*width)*4 + 1 ];
				*pDat++ = data[ (x+y*width)*4 + 2 ];
			}
		}
	}

	FILE *fp=fopen( "snap.bmp","wb" );
	fwrite( pBuf,filesize,1,fp );
	fclose(fp);
}


main( int argc,char *argv[] )
{
	
	using namespace raytrace;
	
	Render render;
	Scene  scene;
	View   view;
	SceneLoader loader( &render,&scene,&view );
	
	int scrnwidth=128*4,scrnheight=128*4;
	render.SetScreenSize( scrnwidth,scrnheight );
	
	char *fname="test.scene";
	if( argc==2 ) {
		fname=argv[1];
	}
	
	if( ! loader.LoadText( fname ) )
	{
		fprintf( stderr,"cannot open %s.\n",fname );
		return 0;
	}
	


#ifdef WIN32
	DWORD t0=GetTickCount();
	render.Rendering( &scene,&view );
	{
		FILE *fp=fopen("debug.txt","wt");
		fprintf( fp,"%fsec\n",0.001*(GetTickCount()-t0) );
		fclose(fp);
	}
#else
	render.Rendering( &scene,&view );
#endif
	
	SaveBmp( render.GetScreen(), render.GetWidth(),render.GetHeight(),32 );

	return 0;
}
