

#include "raytrace.h"


using namespace raytrace;

#ifdef	FLOATISDOUBLE
const double MARGIN=1e-7;
#else
const double MARGIN=1e-4;
#endif

#ifndef	BYTE
typedef	unsigned char BYTE;
#endif

template<class T>
inline T setrange( T x,T bottom,T top ) {
	if( x<bottom ) x=bottom;
	else if ( x>top ) x=top;
	return x;
}


Render::Render()
{
}

Render::~Render()
{
}


bool Render::SetScreenSize( int width,int height )
{
	this->width=width;
	this->height=height;

	pScreenData = new char[width*height*4];
	
	return true;
}

void Render::Rendering( Scene* pScene,View* pView )
{
	
	for( int y=0;y<height;y++ )
    {
		for( int x=0;x<width;x++ )
		{
			vector3 raydir=vector3(
				 pView->projpoint
				+pView->projx * (x-width/2)/width
				+pView->projy * (y-height/2)/height
				-pView->eyepoint
				).unit();
			vector3 raystart=pView->eyepoint;

			color c;
			RayTrace( &c, pScene, raystart,raydir, 1.0 ); 
			((BYTE*) pScreenData)[ (x+y*width)*4+2 ] = setrange<int>( c.GetR()*256,0,255 );
			((BYTE*) pScreenData)[ (x+y*width)*4+1 ] = setrange<int>( c.GetG()*256,0,255 );
			((BYTE*) pScreenData)[ (x+y*width)*4+0 ] = setrange<int>( c.GetB()*256,0,255 );
			
		}
		printf("%3d%%\x0d",100*y/height);
    }
	
	printf("\n\n");
}

void Render::RayTrace( color* retcolor, Scene* pScene, vector3 start,vector3 dir, double ratio, Object* transmitobj )
{
	const double RATIOMIN=0.01;
	int objmax=pScene->object.size();

	if( ratio<RATIOMIN ) {
		*retcolor=0.0;
		return;
	}
	
	
	// get most near object
	double mindistance=-1;
	vector3 ret[2];
	vector3 crosspos,normal;
	Object* object=0;
	while( --objmax>=0 )
    {
		vector3 ret[2];
		double   distance;
		if( 0==pScene->object[objmax]->DetectCross( ret,&distance, start+dir*MARGIN,dir ) ) continue;
		if( distance<mindistance || mindistance<0 )
		{
			mindistance=distance;
			object=pScene->object[objmax];
			crosspos=ret[0];
			normal=ret[1];
		}
    }

	
	if( object==0 ){
		*retcolor=color(0,0,0);
		return;
	}
	
	color c=0.0;
	Material m=*object->GetMaterial( crosspos );
	
	// diffuse,specular
	int lightmax=pScene->light.size();
	while( --lightmax>=0 )
    {
		Light* pLight=pScene->light[lightmax];
		
		int objmax=pScene->object.size();
		Object* object;
		double alpha=1.0;
		while( --objmax>=0 )
		{
			vector3 ret[2];
			vector3 dir=(pLight->pos-crosspos).unit();
			double  distance;
			object=pScene->object[objmax];
			if( object->DetectCross( ret,&distance,crosspos+dir*MARGIN,dir ) > 0 ) {
				Material* m=object->GetMaterial( ret[0] );
				alpha *= m->GetTransparency();
				if( alpha<MARGIN ) goto nextlight;
			}
		}

		{
		double damp = (1.0-pLight->damping) + pLight->damping/(pLight->pos-crosspos).norm();
		vector3 in = (pLight->pos-crosspos).unit();
		double costheta = in*normal,
		       cosgunma = -(in-2*in*normal*normal)*(start-crosspos).unit();
		
		// spot light
		double spot;
		if( pLight->ndir!=0 ) {
			spot=-in*pLight->dir;
			if( spot>pLight->dircos ) spot=1; else spot=1.0-(pLight->dircos-spot)*(pLight->ndir);
		} else spot=1;
		if( spot<0 ) spot=0; else if( spot>1 ) spot=1;

		// diffuse
		if( costheta>0 )
			c += pLight->c * m.GetDiffuse() * costheta *alpha *damp *spot;
		// specular
		if( cosgunma>0 )
			c += pLight->c * m.GetSpecular() * costheta * pow( cosgunma,m.GetNspecular() ) *alpha *damp *spot;
		}
		
nextlight:;
    }
	
	// ambient
	{
		color ambient=m.GetAmbient();
		if( c.GetR()<ambient.GetR() ) c.SetR( ambient.GetR() );
		if( c.GetG()<ambient.GetG() ) c.SetG( ambient.GetG() );
		if( c.GetB()<ambient.GetB() ) c.SetB( ambient.GetB() );
	}

	double kreflect=m.GetReflect();

	// 
	color  opa=1;
	if( m.GetTransparency()*ratio>=RATIOMIN )
    {
		vector3 indir  = (crosspos-start).unit();
		vector3 indirt = indir - indir*normal*normal;	// ˃xNg@菜
		double sin1 = indirt.distance();
		double a1   = asin( sin1 );						// ˊp
		double n1,n2;
		if( transmitobj==object ) { // ̂oƂ
			n1 = m.GetRefract();
			n2 = 1;
			opa=color(1,1,1)-m.GetOpaque()*pow(mindistance,0.3);
			opa.chk();
		} else {					// ̂ɓƂ
			n2 = m.GetRefract();
			//if( transmitobj ) n1=transmitobj->GetMaterial(crosspos)->GetRefract();	else
			n1=1;
		}

		double sin2 = sin1*n1/n2;
		if( sin2>1.0 ) {
			sin2=1.0;
		}
		double a2   = asin( sin2 );						// ܊p
		double k;

		vector3 dir = -cos(a2)*normal + sin2*indirt.unit();				// ܌̃xNg
		if( a1+a2 != 0 ) {
			k = 0.5*( power2( sin(a1-a2)/sin(a1+a2) ) + power2( tan(a1-a2)/tan(a1+a2) ) );	// ˂銄
		} else {
			k = 0;
		}
		if( k<1.0 ) {
			color tmp;
			RayTrace( &tmp, pScene, crosspos+dir*MARGIN,dir, m.GetTransparency()*(1.0-k)*ratio, object );
			c += m.GetTransparency() * tmp * (1.0-k);
		} else k=1;
		kreflect += k;
	}

	// 
	if( kreflect*ratio>=RATIOMIN )
    {
		if( kreflect>0.99 ) kreflect=0.99;
		vector3 dir = (crosspos-start) - 2*(crosspos-start)*normal*normal;
		dir.normalize();
		color tmp;
		RayTrace( &tmp, pScene, crosspos+dir*MARGIN,dir, kreflect*ratio, object );
		c += kreflect * tmp;
		
    }


	*retcolor=c*opa;
	return;
	
}


// ================================================================================


int ObjectCSG::DetectCross( vector3* ret,double* distance,vector3 start,vector3 dir )
{
	if( type==CSG_OR ) {
		int r1,r2;
		vector3 ret1[2],ret2[2];
		double finaldis=0;
		double distance1,distance2;
		
		while(true)
		{
			r1=c1->DetectCross( ret1,&distance1,start,dir );
			r2=c2->DetectCross( ret2,&distance2,start,dir );
			
			if( r1==0 && r2==0 ) return 0;
			
			if( r1>0 && r2==0 ) {
				ret[0]=ret1[0];ret[1]=ret1[1];
				*distance=finaldis+distance1;
				return 1;
			}
			if( r2>0 && r1==0 ) {
				ret[0]=ret2[0];ret[1]=ret2[1];
				*distance=finaldis+distance2;
				return 1;
			}
			
			if( distance1<distance2 ) {
				if( c2->DetectInside( ret1[0] ) ) {
					start += dir*(distance1+MARGIN);
					finaldis+=distance1;
					continue;
				}
				ret[0]=ret1[0];ret[1]=ret1[1];
				if( ret[1]*dir>0 ) ret[1]=-ret[1];
				*distance=finaldis+distance1;
				return 1;
			} else {
				if( c1->DetectInside( ret2[0] ) ) {
					start += dir*(distance2+MARGIN);
					finaldis+=distance2;
					continue;
				}
				ret[0]=ret2[0];ret[1]=ret2[1];
				if( ret[1]*dir>0 ) ret[1]=-ret[1];
				*distance=finaldis+distance2;
				return 1;
			}
		}
	}
	
	if( type==CSG_AND ) {
		int r1,r2;
		vector3 ret1[2],ret2[2];
		double finaldis=0;
		double distance1,distance2;
		
		while(true)
		{
			r1=c1->DetectCross( ret1,&distance1,start,dir );
			r2=c2->DetectCross( ret2,&distance2,start,dir );
			
			if( r1==0 || r2==0 ) return 0;
			
			if( distance1<distance2 ) {
				if( !c2->DetectInside( ret1[0] ) ) {
					start = start+dir*(distance1+MARGIN);
					finaldis+=distance1;
					continue;
				}
				ret[0]=ret1[0];ret[1]=ret1[1];
				if( ret[1]*dir>0 ) ret[1]=-ret[1];
				*distance=finaldis+distance1;
				return 1;
			} else {
				if( !c1->DetectInside( ret2[0] ) ) {
					start = start+dir*(distance2+MARGIN);
					finaldis+=distance2;
					continue;
				}
				ret[0]=ret2[0];ret[1]=ret2[1];
				if( ret[1]*dir>0 ) ret[1]=-ret[1];
				*distance=finaldis+distance2;
				return 1;
			}
		}
	}

	if( type==CSG_SUB ) {
		int r1,r2;
		vector3 ret1[2],ret2[2];
		double finaldis=0;
		double distance1,distance2;
		
		if( c2->DetectInside( start ) ){
			goto inside;
		}

		while(true)
		{
			r1=c1->DetectCross( ret1,&distance1,start,dir );
			if( r1==0 ) return 0;

			r2=c2->DetectCross( ret2,&distance2,start,dir );
			
			if( distance1<distance2 || r2==0 ) {
				ret[0]=ret1[0];ret[1]=ret1[1];
				if( ret[1]*dir>0 ) ret[1]=-ret[1];
				*distance=finaldis+distance1;
				return 1;
			} else {
				start = ret2[0]+dir*MARGIN;
				finaldis+=distance2;
inside:			r2=c2->DetectCross( ret2,&distance2,start,dir );
				if( r2==0 ) return 0;
				if( c1->DetectInside( ret2[0] ) ) {
					ret[0]=ret2[0];ret[1]=ret2[1];
					if( ret[1]*dir>0 ) ret[1]=-ret[1];
					*distance=finaldis+distance2;
					return 1;
				}
				start = start+dir*(distance2+MARGIN);
				finaldis+=distance2;
			}
		}
	}

	// error
	return 0;
}

bool ObjectCSG::DetectInside( vector3 x )
{
	if( type==CSG_OR ) {
		if( c1->DetectInside( x ) ) return true;
		if( c2->DetectInside( x ) ) return true;
		return false;
	}

	if( type==CSG_AND ) {
		return c1->DetectInside( x ) && c2->DetectInside( x );
	}

	if( type==CSG_SUB ) {
		return c1->DetectInside( x ) && ! c2->DetectInside( x );
	}

	// error
	return false;
}

// ================================================================================


TextureBmp::TextureBmp()
{
	data=0;
}

TextureBmp::~TextureBmp()
{
	delete data;
}

bool TextureBmp::LoadBmp( const char * fname )
{
	typedef unsigned short WORD;
	typedef unsigned int   DWORD;
	typedef          long  LONG;
#pragma pack(push,1)
	typedef struct tagBITMAPFILEHEADER {
		WORD    bfType; 
		DWORD   bfSize; 
		WORD    bfReserved1; 
		WORD    bfReserved2; 
		DWORD   bfOffBits; 
	} BITMAPFILEHEADER; 
	typedef struct tagBITMAPINFOHEADER{
		DWORD  biSize; 
		LONG   biWidth; 
		LONG   biHeight; 
		WORD   biPlanes; 
		WORD   biBitCount;
		DWORD  biCompression; 
		DWORD  biSizeImage; 
		LONG   biXPelsPerMeter; 
		LONG   biYPelsPerMeter; 
		DWORD  biClrUsed; 
		DWORD  biClrImportant; 
	} BITMAPINFOHEADER; 
#pragma pack(pop)
	FILE *fp=fopen( fname,"rb" );
	if( !fp ) return false;

	try {
		BITMAPFILEHEADER bfh;
		BITMAPINFOHEADER bih;
		fread( &bfh,sizeof bfh,1,fp );
		fread( &bih,sizeof bih,1,fp );
		if( bfh.bfType!='B'+'M'*256 ) throw;
		if( bih.biBitCount!=24 ) throw;
		width=bih.biWidth;
		height=bih.biHeight;
		data = new BYTE[ width*height*3 ];
		fread( data,width*height*3,1,fp );
		fclose(fp);
	} catch ( void ) {
		fclose(fp);
		return false;
	}

	return true;
}


// ================================================================================

color TextureBmp::GetColor( VAR x,VAR y )
{
	if( x<0 ) x=0; else if( x>1.0 ) x=1.0;
	if( y<0 ) y=0; else if( y>1.0 ) y=1.0;

	int ix=x*width,iy=y*height;
	VAR dx=x*width-ix,dy=y*height-iy;

	int base = (ix+iy*width)*3;
	if( ix==width-1 || iy==height-1 )
		return color( data[base+2],data[base+1],data[base] )*(1.0/255);
	return (   color( data[base          +2],data[base          +1],data[base          ] )*(1.0-dx)*(1.0-dy)
			 + color( data[base+3        +2],data[base+3        +1],data[base+3        ] )*(    dx)*(1.0-dy)
	         + color( data[base+  width*3+2],data[base+  width*3+1],data[base+  width*3] )*(1.0-dx)*(    dy)
	         + color( data[base+3+width*3+2],data[base+3+width*3+1],data[base+3+width*3] )*(    dx)*(    dy) ) * (1.0/255);
}

int ObjectTexture::ExtendCommand( const char* command, void* param1, void* param2 )
{
	if( strcmp( command,"texture" )==0 )
	{
		return SetTexture( (const char *)param1 );
	}

	return ObjectMat::ExtendCommand( command,param1,param2 );
}
