


#ifndef	__RAYTRACE_H
#define	__RAYTRACE_H

#include <vector>
#include "vector3.h"
#include "quaternion.h"

typedef unsigned char BYTE;

template<class T>
inline T power2( T x ) { return x*x; }

// ================================================================================

// --------------------------------------------------------------------------------
class color
{
private:
	VAR r,g,b;
	
public:
	inline color() { r=g=b=0; }
	inline color( VAR a ) { Set(a,a,a); }
	inline color( VAR r,VAR g,VAR b ) { Set(r,g,b); }
	inline VAR GetR() { return r; }
	inline VAR GetG() { return g; }
	inline VAR GetB() { return b; }
	inline void SetR(VAR x) { r=x; }
	inline void SetG(VAR x) { g=x; }
	inline void SetB(VAR x) { b=x; }
	inline void Set( VAR r,VAR g,VAR b ) { this->r=r;this->g=g;this->b=b; }
	
	inline color& operator += ( const color& x ) { r+=x.r;g+=x.g;b+=x.b;return *this; }
	inline color& operator -= ( const color& x ) { r-=x.r;g-=x.g;b-=x.b;return *this; }
	inline friend color operator + ( const color& x,const color& y ) { return color( x.r+y.r,x.g+y.g,x.b+y.b ); }
	inline friend color operator - ( const color& x,const color& y ) { return color( x.r-y.r,x.g-y.g,x.b-y.b ); }
	inline friend color operator * ( const color& x,const color& y ) { return color( x.r*y.r,x.g*y.g,x.b*y.b ); }
	inline friend color operator * ( const color& x,VAR y ) { return color( x.r*y,x.g*y,x.b*y ); }
	inline friend color operator / ( const color& x,VAR y ) { return x*(1.0/y); }

	void chk() { if(r<0) r=0; else if(r>1) r=1; if(g<0) g=0; else if(g>1) g=1; if(b<0) b=0; else if(b>1) b=1; }
	
};

// --------------------------------------------------------------------------------

namespace raytrace
{
	

class Material
{
public:
	Material() { diffuse=1.0;specular=0.5;ambient=0;nspecular=1.0;reflect=0;refract=1.0;transparency=0;opaque=0;}
	Material( color dif,color spe,VAR nspe,VAR ref,VAR transparency,VAR refract,color ambient,color opa ) {
		diffuse=dif;specular=spe;nspecular=nspe;reflect=ref;
		this->ambient=ambient;
		this->transparency=transparency;this->refract=refract;this->opaque=opa;
	}

	void SetDiffuse( color c ) { diffuse=c; }
	void SetSpecular( color c ) { specular=c; }
	void SetSpecularPower( VAR n ) { nspecular=n; }
	void SetReflect( VAR n ) { reflect=n; }
	void SetTransparency( VAR n ) { transparency=n; }
	void SetRefract( VAR n ) { refract=n; }
	void SetAmbient( color c ) { ambient=c; }
	void SetOpaque( color c ) { opaque=c; }

	inline color GetDiffuse() { return diffuse; }
	inline color GetSpecular() { return specular; }
	inline color GetAmbient() { return ambient; }
	inline VAR   GetNspecular() { return nspecular; }
	inline VAR   GetReflect() { return reflect; }
	inline VAR   GetTransparency() { return transparency; }
	inline VAR   GetRefract() { return refract; }
	inline color GetOpaque() { return opaque; }

private:
	color diffuse;
	color specular;
	color ambient;
	VAR nspecular;
	VAR reflect;		// ˗
	VAR transparency;	// x
	VAR refract;		// ܗ
						//  KX(yNE)   1.515
						//  KX(dtg)   1.752
						//  _CAh         2.4173
						//                     1.3330
						//  C                 1.000
	color opaque;
};


// --------------------------------------------------------------------------------

class Texture
{
public:
	Texture();
	virtual ~Texture();
	virtual void GetColor( color* c,vector3 x ) = 0;
};

class TextureBmp
{
public:
	TextureBmp();
	~TextureBmp();
	TextureBmp( const char * name );
	bool LoadBmp( const char * fname );
	color GetColor( VAR x,VAR y );		// x,y͐Kꂽlw
private:
	BYTE *data;
	int width,height;
};

// --------------------------------------------------------------------------------

class Light
{
public:
	Light( color c,vector3 pos,VAR d=0,vector3 dir=vector3(1,1,1),VAR ndir=0,VAR dircos=0 ) {
		this->c=c;this->pos=pos;damping=d;this->dir=dir.unit();this->ndir=ndir;this->dircos=dircos;
	}
	color c;
	vector3 pos;
	VAR   damping;		// (0...1) 0Ō
	vector3 dir;
	VAR   ndir,dircos;
};

// --------------------------------------------------------------------------------

class Object
{
public:
	Object() {}
	virtual ~Object() {}
public:
	// 0=kousa nasi 1=kousa ari
	// output : ret[0]=cross point
	//          ret[1]=cross point normal?
	virtual int DetectCross( vector3* ret,VAR* distance,vector3 start,vector3 dir ) = 0;

	// e|ւ̃|C^Ԃꍇ̂ŁCPvIɎgꍇ̓CX^XRs[Ďg
	virtual Material* GetMaterial( vector3 globalpos ) = 0;
	virtual bool DetectInside( vector3 x ) = 0;

	// input : command="texture" param1=gfxfilename param2=wrapping method
	virtual int ExtendCommand( const char* command, void* param1=0, void* param2=0 ) { return -1; }
};



class ObjectMat : public Object
{
protected:
	Material material;
	
	Material* GetMaterial( vector3 ) {
		return &material;
	}
	int ExtendCommand( const char* command, void* param1, void* param2 ) {
		if( 0 == strcmp( command,"material" ) ) {
			SetMaterial( *((Material*)param1) );
			return 1;
		}
		return Object::ExtendCommand( command,param1,param2 );
	}
	
public:
	ObjectMat() {}
	void SetMaterial( Material& m ) {
		material=m;
	}
};

class ObjectTexture : public ObjectMat
{
public:
	ObjectTexture() {
		pTexture=0;
	}
	bool SetTexture( const char * name ) {
		delete pTexture;
		pTexture = new TextureBmp;
		return pTexture->LoadBmp( name );
	}

protected:
//	Material* GetMaterial( vector3 x );
	TextureBmp* GetTexture() { return pTexture; }
	Material tmpm;
private:
	TextureBmp *pTexture;
	int ExtendCommand( const char* command, void* param1, void* param2 );
};

class Sphere : public ObjectTexture
{
private:
	vector3 pos;
	VAR radius;
	
public:
	Sphere() {
		pos=0.0;
		radius=0.1;
	}
	Sphere( vector3 pos,VAR r ) {
		this->pos=pos;
		radius=r;
	}
	~Sphere() {}
	
protected:
	int DetectCross( vector3* ret,VAR* distance,vector3 start,vector3 dir )
	{
		VAR a  = dir*dir,
		      b  = 2*(start-pos)*dir,
		      c  = (start-pos)*(start-pos)-radius*radius;
		
		VAR t0 = b*b-4.0*a*c;
		if( t0<0 ) return 0;
		
		VAR t1 = -b*0.5/a,
		      t2 = sqrt( t0 )*0.5/a;
		VAR d;
		if( t1-t2>0 ) d=t1-t2; else d=t1+t2; // start=̓ ̂߂̏
		ret[0] = start + dir * d;
		ret[1] = ( ret[0]-pos ).unit();
		if( ret[1]*dir>0 ) ret[1]=-ret[1];		// ɕ\ɂ
		*distance = d;
		if( *distance<0 ) return 0;
		return 1;
	}

	bool DetectInside( vector3 x )
	{
		return (pos-x)*(pos-x)<=radius;
	}

	Material* GetMaterial( vector3 x )
	{
		vector3 p=x-pos;
		TextureBmp *pTex = GetTexture();
		if( pTex ) {
			color c = pTex->GetColor(0.5+0.5*p[0]/radius,0.5+0.5*p[1]/radius );
			tmpm=material;
			tmpm.SetDiffuse( c*material.GetDiffuse() );
			tmpm.SetAmbient( c*material.GetAmbient() );
			return &tmpm;
		}
		return ObjectMat::GetMaterial(x);
	}
	
	
};



class QuadricSurface : public virtual ObjectTexture
{
private:
	VAR a,b,c,d,e,f,g,h,i,j;
	
public:
	QuadricSurface() {
	  a=b=c=d=e=f=g=h=i=j=0;
	}
	QuadricSurface( VAR *p ) {
		a=p[0];b=p[1];c=p[2];
		d=p[3];e=p[4];f=p[5];
		g=p[6];h=p[7];i=p[8];
		j=p[9];
	}
	QuadricSurface( VAR p1,VAR p2,VAR p3,VAR p4,VAR p5,VAR p6,VAR p7,VAR p8,VAR p9,VAR p10 ) {
	  a=p1;	  b=p2;	  c=p3;
	  d=p4;	  e=p5;	  f=p6;
	  g=p7;	  h=p8;	  i=p9;
	  j=p10;
	}
	
	~QuadricSurface() {
	}
       
protected:
	int DetectCross( vector3* ret,VAR* distance,vector3 start,vector3 dir )
	{
		VAR A = a*dir[0]*dir[0] + b*dir[1]*dir[1] + c*dir[2]*dir[2] + d*dir[0]*dir[1] + e*dir[1]*dir[2] + f*dir[2]*dir[0];

		if( A==0 ) return 0;

		VAR B = 2*( a*start[0]*dir[0] + b*start[1]*dir[1] + c*start[2]*dir[2] )
			   + d*( start[0]*dir[0] + start[1]*dir[1] ) 
			   + e*( start[1]*dir[1] + start[2]*dir[2] )
			   + f*( start[2]*dir[2] + start[0]*dir[0] )
			   + g*dir[0] + h*dir[1] + i*dir[2],
		    C =  a*start[0]*start[0] + b*start[1]*start[1] + c*start[2]*start[2]
			   + d*start[0]*start[1] + e*start[1]*start[2] + f*start[2]*start[0]
			   + g*start[0] + h*start[1] + i*start[2] + j;

		VAR t0 = B*B-4.0*A*C;
		if( t0<0 ) return 0;

		VAR t1 = -B*0.5/A,
		    t2 = sqrt( t0 )*0.5/A;
		VAR dis;
		if( t1-t2>0 ) dis=t1-t2; else dis=t1+t2;
		vector3 x=start + dir * dis;
		ret[0] = x;
		ret[1] = vector3( 2*a*x[0]                                + d     *x[1]               + f*x[2]      + g                           ,
			                            2*b   *x[1]               + d*x[0]      + e     *x[2]                        + h                  ,
			                                          2*c   *x[2]               + e*x[1]      + f     *x[0]                   + i          )
			.unit();
		if( ret[1]*dir>0 ) ret[1]=-ret[1];		// ɕ\ɂ
		*distance = dis;
		if( *distance<0 ) return 0;
		return 1;

	}
	
	bool DetectInside( vector3 x )
	{
		return (a*x[0]*x[0] + b*x[1]*x[1] + c*x[2]*x[2] + d*x[0]*x[1] + e*x[1]*x[2] + f*x[2]*x[0] + g*x[0] + h*x[1] + i*x[2] + j)
		       <= 0;
	}
};


enum CSGTYPE {
	CSG_OR=0,CSG_AND,CSG_SUB
};
class ObjectCSG : public Object
{
	Object *c1,*c2;
public:
	CSGTYPE type;

	ObjectCSG() { type=CSG_OR; }

	void SetType( CSGTYPE n ) { type=n; }
	void SetObject( Object *p1,Object *p2 ) {
		c1=p1;c2=p2;
	}


	int DetectCross( vector3* ret,VAR* distance,vector3 start,vector3 dir );

	Material* GetMaterial( vector3 x ) {
		return c1->GetMaterial( x );
	}

	bool DetectInside( vector3 x );

};


class Floor : public ObjectTexture
{
public:
	Floor( vector3 pos,vector3 normal,color c1,color c2,VAR size ) {
		this->c1=c1;
		this->c2=c2;
		if( size==0 ) this->size=0; else this->size=1.0/size;
		this->pos=pos;
		this->normal=normal.unit();
	}
protected:
	color c1,c2;
	VAR size;
	vector3 pos,normal;

protected:
	int DetectCross( vector3* ret,VAR* distance,vector3 start,vector3 dir )
	{
		VAR a=dir*normal;
		if( a<0.000001 && a>-0.000001 ) return 0; // 0ŊȂ悤
		VAR t=-(start-pos)*normal / a;
		if( t<0 ) return 0;
		ret[0]=start+dir*t;
		ret[1]=normal;
		if( ret[1]*dir>0 ) ret[1]=-ret[1];		// ɕ\ɂ
		*distance=t;
		return 1;
	}

	bool DetectInside( vector3 x )
	{
		return (x-pos)*normal >= 0;
	}


	Material* GetMaterial( vector3 p ) {

		TextureBmp *pTex = GetTexture();
		if( pTex ) {
			double u=0.5*p[0]/size+4096,v=0.5*p[1]/size+4096;
			color c = pTex->GetColor( u-((int)u),v-((int)v) );
			tmpm=material;
			tmpm.SetDiffuse( c*material.GetDiffuse() );
			tmpm.SetAmbient( c*material.GetAmbient() );
		} else {
			color c;
			if( sin(p[0]*size)*sin(p[1]*size)<0 ) c=c1; else c=c2;
			tmpm=material;
			tmpm.SetDiffuse( c );
			tmpm.SetAmbient( c*material.GetAmbient() );
		}
		return &tmpm;
	}
};

class SimpleMesh : public ObjectTexture
{
private:
	std::vector<vector3>   vertices;
	std::vector<vector3>   verticesR;
	std::vector<int>       indices;
	class Index {
	public:
		int start;
		int len;
		Index( int start,int len ) { this->start=start;this->len=len; }
	};
	std::vector<Index>      index2;
	vector3 pos;
	quaternion q;
	matrix3x3 R;
	vector3 meshsize;

public:
	SimpleMesh() {
		pos=0;q=1;
		meshsize=1;
	}

	void SetPos( vector3 pos ) {
		this->pos=pos;
		RecalcPos();
	}
	void SetQ( quaternion q ) {
		this->q=q;
		RecalcPos();
	}
	void SetSize( vector3 v ) {
		meshsize=v;
	}

	void RecalcPos() {
		q.normalize();
		R = q.rotation();
		for( int i=vertices.size()-1;i>=0;i-- )
		{
			verticesR[i] = meshsize.indmul( R*vertices[i] ) + pos;
		}
	}

	bool loadMQO( const char* filename, const char* objname, VAR size );

	void SetVertices( int size,vector3 *v )
	{
		vertices.clear();
		vertices.resize( size );
		verticesR.resize( vertices.size() );
		for( int i=0;i<size;i++ )
			vertices[i]=v[i];
	}

	void AddSurface( int vnum,int *index )
	{
		index2.push_back( Index( indices.size(),vnum ) );
		for( int i=0;i<vnum;i++ ) indices.push_back( index[i] );
	}


	void SetBox( vector3 pos,quaternion q,VAR width,VAR height,VAR depth )
	{
		this->pos=pos;
		this->q=q;

		VAR x=width/2,y=height/2,z=depth/2;
		q=1.0;

		// ʂ̒_
		vertices.push_back( vector3( -x,-y,z ) );
		vertices.push_back( vector3( -x, y,z ) );
		vertices.push_back( vector3(  x, y,z ) );
		vertices.push_back( vector3(  x,-y,z ) );

		// ʂ̒_
		vertices.push_back( vector3( -x,-y,-z ) );
		vertices.push_back( vector3( -x, y,-z ) );
		vertices.push_back( vector3(  x, y,-z ) );
		vertices.push_back( vector3(  x,-y,-z ) );

		verticesR.resize( vertices.size() );

		// ʒ`
		// EnŊO\
		indices.push_back( 3 ); indices.push_back( 2 ); indices.push_back( 1 ); indices.push_back( 0 );	// 
		indices.push_back( 4 ); indices.push_back( 5 ); indices.push_back( 6 ); indices.push_back( 7 );	// 
		indices.push_back( 3 ); indices.push_back( 7 ); indices.push_back( 6 ); indices.push_back( 2 );	// O
		indices.push_back( 0 ); indices.push_back( 1 ); indices.push_back( 5 ); indices.push_back( 4 );	// 
		indices.push_back( 0 ); indices.push_back( 4 ); indices.push_back( 7 ); indices.push_back( 3 );	// 
		indices.push_back( 1 ); indices.push_back( 2 ); indices.push_back( 6 ); indices.push_back( 5 );	// E

		for( int i=0;i<6;i++ ) index2.push_back( Index( 4*i,4 ) );
		RecalcPos();
	}

	// Op
	void SetTriangularPyramid( vector3 pos,quaternion q,VAR width,VAR height )
	{
		this->pos=pos;
		this->q=q;

		// _
		vertices.push_back( vector3( -width/2,-width  /3,0 ) );
		vertices.push_back( vector3(  width/2,-width  /3,0 ) );
		vertices.push_back( vector3(  0,       width*2/3,0 ) );
		vertices.push_back( vector3(  0,0,height ) );

		verticesR.resize( vertices.size() );

		// ʒ`
		// EnŊO\
		indices.push_back( 0 ); indices.push_back( 2 ); indices.push_back( 1 ); // 
		indices.push_back( 0 ); indices.push_back( 1 ); indices.push_back( 3 ); // 
		indices.push_back( 1 ); indices.push_back( 2 ); indices.push_back( 3 ); // 
		indices.push_back( 0 ); indices.push_back( 3 ); indices.push_back( 2 ); // 

		for( int i=0;i<4;i++ ) index2.push_back( Index( 3*i,3 ) );
		RecalcPos();
	}

	int DetectCross( vector3* ret,VAR* distance,vector3 start,vector3 dir )
	{
		VAR mindistance=-1;
		int minindex2=-1;
		vector3 minnormal;
		vector3 minx;
		for( int i=index2.size()-1;i>=0;i-- )
		{
			vector3 n = ( verticesR[indices[index2[i].start+1]]-verticesR[indices[index2[i].start]] )
				  .cross( verticesR[indices[index2[i].start+2]]-verticesR[indices[index2[i].start]] ).unit();
			VAR a = dir*n;
			if( a==0 ) continue;
			VAR t = (verticesR[indices[index2[i].start]]-start)*n / a;
			if( t<0 ) { continue; }
			vector3 x = start+dir*t;
			
			// ʕʂ̒m߂
			for( int j=index2[i].len-1;j>=0;j-- )
			{
				vector3 v;
				v  =  ( verticesR[indices[index2[i].start+(j+1)%index2[i].len]]-verticesR[indices[index2[i].start+j]] )
				.cross( x                                                      -verticesR[indices[index2[i].start+j]] );
				if( v*n<0 ) goto index2_continue;
			}

			if( t<mindistance || mindistance<0 )
			{
				minindex2=i;
				mindistance = t;
				minnormal=n;
				minx=x;
			}
index2_continue:;
		}

		if( minindex2>=0 ) {
			ret[0]=minx;
			if( minnormal*dir>0 ) ret[1]=-minnormal;
			else 			      ret[1]= minnormal;
			*distance=mindistance;
			return 1;
		} else {
			return 0;
		}
	}


	bool DetectInside( vector3 x )
	{
		for( int i=index2.size()-1;i>=0;i-- )
		{
			vector3 n = ( verticesR[indices[index2[i].start+1]]-verticesR[indices[index2[i].start]] )
				  .cross( verticesR[indices[index2[i].start+2]]-verticesR[indices[index2[i].start]] );
			VAR a = (x-verticesR[indices[index2[i].start]])*n;
			if( a>0 ) return false;
		}
		return true;
	}

};


// --------------------------------------------------------------------------------

class Scene
{
public:
	std::vector<Light*>  light;
	std::vector<Object*> object;
	
public:
	void AddLight( Light* l ) { light.push_back(l); }
	void AddObject( Object* o ) { object.push_back(o); }
};

// --------------------------------------------------------------------------------

class View
{
public:
	void SetEyePoint( vector3& p ) { eyepoint=p; }
	void SetProjection( vector3& point,vector3& x ) {
		projpoint=point;
		projx=x;
		projy=(eyepoint-point).unit().cross(x);
	}
	void SetView( vector3& eye,vector3& dir,vector3& up,VAR x,VAR y )
	{
		eyepoint = eye;
		projpoint = eye + dir;
		projx = up.cross(dir).unit()*x;
		projy = dir.cross(projx).unit()*y;
	}
	
	vector3& GetEyePoint() const;
	vector3& GetProjPoint() const;
	vector3& GetProjx() const;
	vector3& GetProjy() const;

	vector3 eyepoint;
	
	// e
	vector3 projpoint;
	vector3 projx,projy;
};


// --------------------------------------------------------------------------------

// _
// 32bppp
class Render
{
public:
	Render();
	~Render();
	
	bool SetScreenSize( int width,int height );
	void Rendering( Scene* pScene,View* pView );
	void RayTrace( color* retcolor, Scene* pScene, vector3 start,vector3 dir, VAR ratio, Object* transmitobj=0 );
	void* GetScreen() { return pScreenData; }

	int GetWidth() { return width; }
	int GetHeight() { return height; }
	
private:
	int width,height;
	void* pScreenData;
};







};

#endif //__RAYTRACE_H








